@@ -95,7 +95,7 @@ way for wrapping up the functions.
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #
 
-__revision__ = "/home/scons/scons/branch.0/branch.96/baseline/src/engine/SCons/Action.py 0.96.91.D001 2005/09/08 09:14:36 knight"
+__revision__ = "/home/scons/scons/branch.0/branch.96/baseline/src/engine/SCons/Action.py 0.96.92.D001 2006/04/10 23:13:27 knight"
 
 import dis
 import os
@@ -409,8 +409,8 @@ class CommandAction(_ActionAction):
             if c:
                 return c
         cmd_list, ignore, silent = self.process(target, source, env)
-	if silent:
-	    return ''
+        if silent:
+            return ''
         return _string_from_cmd_list(cmd_list[0])
 
     def execute(self, target, source, env):
@@ -422,7 +422,8 @@ class CommandAction(_ActionAction):
         handle lists of commands, even though that's not how we use it
         externally.
         """
-        from SCons.Util import is_String, is_List, flatten, escape_list
+        from SCons.Subst import escape_list
+        from SCons.Util import is_String, is_List, flatten
 
         try:
             shell = env['SHELL']
@@ -461,7 +462,7 @@ class CommandAction(_ActionAction):
                     # reasonable for just about everything else:
                     ENV[key] = str(value)
 
-	cmd_list, ignore, silent = self.process(target, map(rfile, source), env)
+        cmd_list, ignore, silent = self.process(target, map(rfile, source), env)
 
         # Use len() to filter out any "command" that's zero-length.
         for cmd_line in filter(len, cmd_list):
@@ -478,12 +479,13 @@ class CommandAction(_ActionAction):
         This strips $(-$) and everything in between the string,
         since those parts don't affect signatures.
         """
+        from SCons.Subst import SUBST_SIG
         cmd = self.cmd_list
         if SCons.Util.is_List(cmd):
             cmd = string.join(map(str, cmd))
         else:
             cmd = str(cmd)
-        return env.subst_target_source(cmd, SCons.Util.SUBST_SIG, target, source)
+        return env.subst_target_source(cmd, SUBST_SIG, target, source)
 
 class CommandGeneratorAction(ActionBase):
     """Class for command-generator actions."""
@@ -643,7 +645,15 @@ class FunctionAction(_ActionAction):
             result = self.execfunction(target=target, source=rsources, env=env)
         except EnvironmentError, e:
             # If an IOError/OSError happens, raise a BuildError.
-            raise SCons.Errors.BuildError(node=target, errstr=e.strerror)
+            # Report the name of the file or directory that caused the
+            # error, which might be different from the target being built
+            # (for example, failure to create the directory in which the
+            # target file will appear).
+            try: filename = e.filename
+            except AttributeError: filename = None
+            raise SCons.Errors.BuildError(node=target,
+                                          errstr=e.strerror,
+                                          filename=filename)
         return result
 
     def get_contents(self, target, source, env):
@@ -657,22 +667,33 @@ class FunctionAction(_ActionAction):
         So we remove the line number byte codes to prevent
         recompilations from moving a Python function.
         """
+        execfunction = self.execfunction
         try:
-            # "self.execfunction" is a function.
-            contents = str(self.execfunction.func_code.co_code)
+            # Test if execfunction is a function.
+            code = execfunction.func_code.co_code
         except AttributeError:
-            # "self.execfunction" is a callable object.
             try:
-                contents = str(self.execfunction.__call__.im_func.func_code.co_code)
+                # Test if execfunction is a method.
+                code = execfunction.im_func.func_code.co_code
             except AttributeError:
                 try:
-                    # See if execfunction will do the heavy lifting for us.
-                    gc = self.execfunction.get_contents
+                    # Test if execfunction is a callable object.
+                    code = execfunction.__call__.im_func.func_code.co_code
                 except AttributeError:
-                    # This is weird, just do the best we can.
-                    contents = str(self.execfunction)
+                    try:
+                        # See if execfunction will do the heavy lifting for us.
+                        gc = self.execfunction.get_contents
+                    except AttributeError:
+                        # This is weird, just do the best we can.
+                        contents = str(self.execfunction)
+                    else:
+                        contents = gc(target, source, env)
                 else:
-                    contents = gc(target, source, env)
+                    contents = str(code)
+            else:
+                contents = str(code)
+        else:
+            contents = str(code)
         contents = remove_set_lineno_codes(contents)
         return contents + env.subst(string.join(map(lambda v: '${'+v+'}',
                                                      self.varlist)))
