@@ -27,14 +27,16 @@ The main package for the SCons software construction utility.
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #
 
-__revision__ = "/home/scons/scons/branch.0/baseline/src/engine/SCons/__init__.py 0.96.1.D001 2004/08/23 09:55:29 knight"
+__revision__ = "/home/scons/scons/branch.0/branch.96/baseline/src/engine/SCons/__init__.py 0.96.92.D002 2006/04/11 07:39:43 knight"
 
-__version__ = "0.96.1"
+__version__ = "0.96.92"
 
-__build__ = "D001"
+__build__ = "D002"
 
-__buildsys__ = "casablanca"
+__buildsys__ = "roxbury"
 
-__date__ = "2004/08/23 09:55:29"
+__date__ = "2006/04/11 07:39:43"
 
 __developer__ = "knight"
+
+import SCons.Memoize
